/*
 * File         : BV_VideoApp.C
 * Author       : Tomasz Stachowiak
 * Created      : Fri Oct 16 17:16:04 1998
 * Revision     : $Revision$
 * Updated      : $Date$
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

#include <sys/types.h>
#include <sys/time.h>
#include <semaphore.h>

#include "H263.h"
#include "H261.h"
#include "rtp.h"
#include "MainForm.h"
#include "BV.h"
#include "BV_Application.h"
#include "BV_VideoCapThread.h"
#include "BV_Member.h"
#include "BV_Session_Control.h"
#include "BV_VideoApp.h"

BV_VideoApp::BV_VideoApp() : BV_Application(VIDEO_PLUGIN_PORT)
{
  session_ = BV_Session_Control::instance();
  main_form_ = session_->getMainForm();

  state_ = STOP;
  compression_ = H263;
  first = 1;
  frame_no = 0;
  msg_no = 0;

  sem_init(&semH263_, 0, 1);
  sem_init(&semH261_, 0, 1);

  H263initEncoder(176, 144);
  H263initDecoder();

  /* H.261 */ 
  init_codec();
  init_encoder();
  init_decoder(); 

  initcapture(176, 144);
  init_global_dpy();

  
  cap_thread_ = new BV_VideoCapThread(this);
  cap_thread_->start();
  
}

BV_VideoApp::~BV_VideoApp()
{
  cap_thread_->stop();
  sleep(1);
  delete cap_thread_;
}

int BV_VideoApp::process_message(RTPheader *header, char *body, int len)
{
  char name[50];
  int sender = header->ssrc;
  BV_Member *mbr = session_->get_member(sender);
  if(mbr == NULL)
    return -1;

  mbr->get_name(name);

  if(header->type != 2)
    {
      return -1;
    }

  switch (header->pt)
    {
    case H263:
      if(!mbr->display_on_)
	{
	  init_display("", name, (int)((MainForm *)main_form_)->baseWidget(), 
		       mbr->get_internal_id(), 1);
	  mbr->display_on_ = 1;
	}
      displayH263(mbr->get_internal_id(), body, len);
      break;

    case H261:
      if(!mbr->display_on_)
	{
	  init_display("", name, (int)((MainForm *)main_form_)->baseWidget(), 
		       mbr->get_internal_id(), 1);
	  mbr->display_on_ = 1;
	}
      displayH261(mbr->get_internal_id(), body, len);
      break;

    case YUV9:
      if(!mbr->display_on_)
	{
	  init_display("", name, (int)((MainForm *)main_form_)->baseWidget(), 
		       mbr->get_internal_id(), 1);
	  mbr->display_on_ = 1;
	}
      displayYUV9(mbr->get_internal_id(), body, len);
      break;

    case videoTagPlay:
      if(!mbr->display_on_)
	{
	  init_display("", name, (int)((MainForm *)main_form_)->baseWidget(), 
		       mbr->get_internal_id(), 1);
	  mbr->display_on_ = 1;
	}
      break;

    case videoTagStop:
      if(mbr->display_on_)
	{
	  close_display(mbr->get_internal_id());
	  mbr->display_on_ = 0;
	}
      break;

    default:
      break;
    }

  return 0;
}


int BV_VideoApp::displayH263(int mbr, char *data, int len)
{
  VideoPluginFrame *frame;
  struct timeval start, stop;
  unsigned char *dec_frame;
  int lastDecodeTime;
  int size;
  
  gettimeofday(&start);
  frame = (VideoPluginFrame *)data;
  size = len - sizeof(VideoPluginFrame) + 1;

  sem_wait(&semH263_);
  dec_frame = (unsigned char *)H263decodeFrame(mbr, frame->data, size);
  sem_post(&semH263_);

  if(dec_frame != NULL)
    {
      displayYUV411(mbr, (char *)dec_frame);
      free(dec_frame);
    }

  gettimeofday(&stop);
  lastDecodeTime = (stop.tv_sec - start.tv_sec)*1000 
    + (stop.tv_usec - start.tv_usec)/1000;
  
  if(dec_frame == NULL)
    return -1;

  return 0;
}

int BV_VideoApp::displayH261(int mbr, char *data, int len)
{
  VideoPluginFrame *frame;
  struct timeval start, stop;
  unsigned char *dec_frame;
  int lastDecodeTime;
  int size;
  
  gettimeofday(&start);
  frame = (VideoPluginFrame *)data;
  size = len - sizeof(VideoPluginFrame) + 1;
  dec_frame = (unsigned char *)malloc(176*144*3/2);

  /*
  if(h261first)
    {
      decode_frame(frame->data, size, dec_frame, 1, mbr);
      h261first = 0;
    }
    else */
  sem_wait(&semH261_);
  decode_frame(frame->data, size, (char *)dec_frame, 0, mbr);
  sem_post(&semH261_);

  displayYUV411(mbr, (char *)dec_frame);
  free(dec_frame);

  gettimeofday(&stop);
  lastDecodeTime = (stop.tv_sec - start.tv_sec)*1000 
    + (stop.tv_usec - start.tv_usec)/1000;
  
  if(dec_frame == NULL)
    return -1;

  return 0;
}

int BV_VideoApp::displayYUV9(int mbr, char *data, int len)
{
  VideoPluginFrame *frame;
  struct timeval start, stop;
  unsigned char *dec_frame;
  int lastDecodeTime;
  int size;
  
  gettimeofday(&start);

  frame = (VideoPluginFrame *)data;
  size = len - sizeof(VideoPluginFrame) + 1;
  dec_frame = (unsigned char *)malloc(176*144*3/2);
  YVU9toYUV411((unsigned char *)frame->data, dec_frame, &dec_frame[176*144], 
	       &dec_frame[176*144*5/4], 176, 144);
  displayYUV411(mbr, (char *)dec_frame);
  free(dec_frame);

  gettimeofday(&stop);
  lastDecodeTime = (stop.tv_sec - start.tv_sec)*1000 
    + (stop.tv_usec - start.tv_usec)/1000;
  
  if(dec_frame == NULL)
    return -1;

  return 0;
}



int BV_VideoApp::send_video()
{
  unsigned char *frame;
  char buf[10000];
  int size;

  if(state_ == PLAY)
    {
      frame = getFrame(176, 144);
      
      displayYUV411(MAX_MEMBERS, (char *)frame);

      /* coding must be synchronized with decoding !!!! */
      switch(compression_)
	{
	case H263:
	  sem_wait(&semH263_);
	  if(frame_no == 0)
	    {
	      size = H263codeIntraFrame(frame, buf);
	    }
	  else
	    {		
	      size = H263codeFrame(&frame, buf);
	    }
	  sem_post(&semH263_);
	  break;
	  
	case H261:
	  sem_wait(&semH261_);
	  if(frame_no == 0)
	    size = p64EncodeIntraFrame((char *)frame, buf);
	  else
	    size = p64EncodeFrame((char *)frame, buf);
	  sem_post(&semH261_);
	  
	  break;
	}
      
      free(frame);
      
      frame_no = (frame_no + 1) % INTRA_GAP;

      send_video_data(compression_, buf, size);
      
      /*sleep();*/
    }
  else
    sleep(1);

  return 0;
}

void BV_VideoApp::play()
{
  if(state_ == STOP)
    {
      init_display("", "Mirror", (int)((MainForm *)main_form_)->baseWidget(), 
		   MAX_MEMBERS, 1);
      state_ = PLAY;
    } 
}

void BV_VideoApp::stop()
{
  if(state_ == PLAY)
    {
      state_ = STOP;
      close_display(MAX_MEMBERS);
    }
}

int BV_VideoApp::send_video_data(int tag, char *data, int len)
{
  VideoPluginFrame *frame;
  int err;

  frame = (VideoPluginFrame *)malloc(sizeof(VideoPluginFrame) + len - 1);
   if (!frame)
   {
     fprintf(stderr, "Message allocation error\n");
     return -1;
   }

  if (len <= 0)
    return -2;

  msg_no = (msg_no + 1) % 256;
  frame->num = (unsigned char)msg_no;
  gettimeofday(&frame->time);
  memcpy(frame->data, data, len);

  err = send_data(compression_, (char *)frame, sizeof(VideoPluginFrame)+len-1);
  
  free(frame);

  return err;
}

void BV_VideoApp::update_options(Options *options)
{
  int tmp_state = state_;
  state_ = STOP;
  if(tmp_state == PLAY)
    sleep(1);

  first = 1;
  frame_no = 0;

  switch(options->connection)
    {
    case _modem28:
      compression_ = H263;
      break;

    case _isdn:
      compression_ = H261;
      break;
      
    case _isdn2:
      compression_ = H261;
      break;

    case _lan:
      compression_ = H261;
      break;

    default:
      compression_ = H263;
      break;
    }

  state_ = tmp_state;
}
